import json
import os
from urllib.parse import urlsplit
from opencensus.trace import config_integration
from opencensus.ext.zipkin.trace_exporter import ZipkinExporter
from opencensus.ext.flask.flask_middleware import FlaskMiddleware
from opencensus.trace.samplers import AlwaysOnSampler, AlwaysOffSampler, \
    ProbabilitySampler
from opencensus.trace.propagation.b3_format import B3FormatPropagator

tracing_config_file = os.getenv('TRACING_CONFIG', '/etc/tracing/client.conf')


def setup_tracing(app, service_name):
    try:
        with open(tracing_config_file) as fd:
            config = json.load(fd)
    except (OSError, IOError):
        return
    if 'report_url' not in config:
        return

    # Configure the Zipkin exporter.
    report_url = urlsplit(config['report_url'])
    report_host, report_port = report_url.netloc.split(':')
    trace_exporter = ZipkinExporter(
        service_name=service_name,
        host_name=report_host,
        port=report_port,
        endpoint=report_url.path,
        protocol=report_url.scheme,
    )

    # Configure the sampler.
    sample_pct = config.get('sample', 'always')
    if sample_pct == 'always':
        trace_sampler = AlwaysOnSampler()
    elif sample_pct == 'never':
        trace_sampler = AlwaysOffSampler()
    else:
        trace_sampler = ProbabilitySampler(float(sample_pct))

    # We use the B3 (Zipkin) propagator because it seems to be the
    # default in the Go implementation.
    propagator = B3FormatPropagator()

    # Set up the OpenCensus Flask middleware.
    middleware = FlaskMiddleware(
        app,
        sampler=trace_sampler,
        exporter=trace_exporter,
        propagator=propagator,
        excludelist_paths=['static'],
    )

    # Patch the 'requests' module so that HTTP clients propagate the
    # trace ID. Note that this is mostly useless, as the requests
    # integration only patches the high-level API (get, post) and does
    # not touch the low-level Session interface that we're using for
    # our own RPC calls.
    config_integration.trace_integrations(['requests'])

    app.logger.info('configured request tracing with service=%s',
                    service_name)
